-- I could make a deprecative joke about the state of my wrist but I decided against it.
-- (Psycho Cut script)

local ENERGY_CONSUMPTION = 18

local wpn2color = {
	[WEP_GRENADE] = SKINCOLOR_BROWN,
	[WEP_AUTO] = SKINCOLOR_NEONGREEN,
	[WEP_EXPLODE] = SKINCOLOR_BLACK,
	[WEP_SCATTER] = SKINCOLOR_PURPLE,
	[WEP_RAIL] = SKINCOLOR_SILVER,
	[WEP_BOUNCE] = SKINCOLOR_ORANGE,
}

local wpn2rw = {
	[WEP_GRENADE] = RW_GRENADE,
	[WEP_AUTO] = RW_AUTO,
	[WEP_EXPLODE] = RW_EXPLODE,
	[WEP_SCATTER] = RW_SCATTER,
	[WEP_RAIL] = RW_RAIL,
	[WEP_BOUNCE] = RW_BOUNCE,
}

-- fires cuts, applying stuff for weapons and consuming them.
local function silv_spawnCut(p)
	local mo = p.mo

	local wpn = not (mo.silv_firenormal) and mo.silv_savewpn or 0
	local color = wpn2color[wpn] or SKINCOLOR_TEAL

	if p.powers[pw_infinityring]
		p.powers[pw_infinityring] = $-1
	else
		mo.silv_energy = $-ENERGY_CONSUMPTION
	end

	if not mo.silv_firenormal
	and wpn2rw[wpn]
		mo.silv_ammo[wpn2rw[wpn]] = max(0, $-1)
		if not mo.silv_ammo[wpn2rw[wpn]]
			mo.silv_fire = nil
		end
	end

	S_StartSound(mo, sfx_pscut)
	local cut = P_SpawnPlayerMissile(mo, MT_PSYCHOCUT)
	if cut and cut.valid
		cut.fuse = TICRATE*10
		cut.target = mo
		cut.color = color
		cut.color2 = color
		cut.height = FRACUNIT*128
		cut.radius = FRACUNIT*32
		cut.weapon = wpn

		if wpn == WEP_AUTO	-- small cut
			cut.scale = mo.scale/2
		elseif wpn == WEP_BOUNCE
			cut.flags = MF_MISSILE|MF_NOGRAVITY|MF_BOUNCE|MF_GRENADEBOUNCE
			cut.fuse = TICRATE*6
			cut.destscale = mo.scale/2
			cut.scalespeed = cut.scale/512
		elseif wpn == WEP_SCATTER
			local savean = mo.angle
			local angs = {savean + ANG1*8, savean - ANG1*8}

			-- Scatters are always horizontal (And slightly smaller)
			cut.scale = mo.scale - (mo.scale/4)
			cut.state = S_PSYCUT2
			if cut and cut.valid
				cut.height = FRACUNIT*32
				cut.radius = FRACUNIT*64
			end

			for i = 1, 2
				mo.angle = angs[i]
				local cut2 = P_SpawnPlayerMissile(mo, MT_PSYCHOCUT)
				cut2.fuse = TICRATE*10
				cut2.target = mo
				cut2.color = color
				cut2.color2 = color
				cut2.scale = mo.scale - (mo.scale/4)
				cut2.height = FRACUNIT*128
				cut2.radius = FRACUNIT*32
				cut2.weapon = wpn
				cut2.state = S_PSYCUT2
				if cut2 and cut2.valid
					cut2.height = FRACUNIT*32
					cut2.radius = FRACUNIT*64
				end
			end
			mo.angle = savean
		elseif wpn == WEP_GRENADE	-- slow moving cut
			cut.fuse = TICRATE*4
			cut.momx, cut.momy, cut.momz = $1/6, $2/6, $3/6
		elseif wpn == WEP_EXPLODE
			cut.explode = true	-- used in a thinker
		elseif wpn == WEP_RAIL
			cut.momx, cut.momy, cut.momz = $1*2, $2*2, $3*2
		end

		if P_RandomRange(0, 1) == 1
		and wpn ~= WEP_SCATTER -- scatters are always horizontal
			cut.state = S_PSYCUT2
			if cut and cut.valid
				cut.height = FRACUNIT*32
				cut.radius = FRACUNIT*64
			end
		end
	end
end

-- returns the max # of cuts and the cooldown
local function silv_getWpnCooldownandTimer(p)
	local wpn = not (p.mo.silv_firenormal) and p.currentweapon or 0
	local cd, t, noreduc = TICRATE, p.powers[pw_super] and 5 or 3

	if wpn == WEP_SCATTER
		t = p.powers[pw_super] and 2 or 1
		noreduc = true
	elseif wpn == WEP_GRENADE
		t = p.powers[pw_super] and 3 or 2
	elseif wpn == WEP_EXPLODE
		t = p.powers[pw_super] and 2 or 1
		cd = TICRATE
		noreduc = true	-- firenormal has no shorter cooldown
	end

	if p.powers[pw_super]
		cd = $/2
	end

	p.mo.silv_savewpn = wpn

	return cd, t
end

-- returns the weapon delay
local function silv_getFireCooldown(p)
	local wpn = not (p.mo.silv_firenormal) and p.currentweapon or 0
	local wd = p.cmd.buttons & BT_ATTACK and TICRATE or TICRATE-(TICRATE/3)

	if wpn == WEP_AUTO
		wd = $/2
	elseif wpn == WEP_SCATTER
		wd = $+($/8)
	elseif wpn == WEP_RAIL
		wd = $+($/2)
	elseif wpn == WEP_EXPLODE
		wd = $*2
	end

	p.mo.silv_savewpn = wpn

	return wd
end

rawset(_G, "silv_psycutHandle", function(p)
	local mo = p.mo
	if not mo return end	-- player is spectating

	if mo.silv_ammo == nil
		mo.silv_ammo = {}
		for i = 0, 15
			mo.silv_ammo[2^i] = 0
		end
	end

	if not G_RingSlingerGametype() return end
	if p.exiting return end
	if p.powers[pw_nocontrol] return end

	if mo.silv_energy <= ENERGY_CONSUMPTION-1 and not p.powers[pw_infinityring]
		mo.silv_fire = nil
		return
	end	-- no rings left / dead

	if p.cmd.buttons & (BT_ATTACK|BT_FIRENORMAL)		-- pressing fire / firenormal (same to me)
	and not (mo.silv_telekinesis and #mo.silv_telekinesis)
	and not (G_TagGametype() and not (p.pflags & PF_TAGIT))
	and mo.silv_alivetime and mo.silv_alivetime > 14
	and mo.health
		if not mo.silv_firehold
		and not mo.silv_fire
		and not mo.silv_firecooldown

			mo.silv_firenormal = p.cmd.buttons & BT_FIRENORMAL
			local maxcooldown, maxfire, noreduc = silv_getWpnCooldownandTimer(p)
			mo.silv_fire = 5*(min(maxfire, p.powers[pw_infinityring] or (mo.silv_energy/ENERGY_CONSUMPTION)))	-- fire up to 3 cuts every 5 tics
			mo.silv_firecooldown = silv_getFireCooldown(p)
			silv_setUseDelay(mo, maxcooldown)
		end
		mo.silv_firehold = true
	else
		mo.silv_firehold = nil
	end

	if mo.silv_fire
		if mo.silv_fire and not (mo.silv_fire%5)	-- spawn one every 5 tics (not on 0!)
		and (p.powers[pw_infinityring] or (mo.silv_energy > ENERGY_CONSUMPTION-1))
			silv_spawnCut(p)
		end
		if mo.silv_fire
			mo.silv_fire = $-1
		end
	elseif mo.silv_firecooldown
		mo.silv_firecooldown = $-1
		if not mo.silv_firecooldown
			mo.silv_firenormal = nil
			mo.silv_savewpn = nil
		end
	end
end)

local function cutCollision(psycut, mo)
	if psycut and psycut.valid
	and mo and mo.valid and (mo.type == MT_PLAYER and mo.flags & MF_SHOOTABLE or mo.flags & (MF_ENEMY|MF_BOSS) and mo.flags & MF_SHOOTABLE)
	and psycut.z+psycut.height >= mo.z and mo.z+mo.height >= psycut.z
	and not mo.silv_grabbed
        if mo.player and psycut.target and psycut.target.player and psycut.target.player.valid
			if friendlyPlayer(mo.player, psycut.target.player)
				return
			end
		end
		P_DamageMobj(mo, psycut, psycut.target)
	end
end

addHook("MobjCollide", cutCollision, MT_PSYCHOCUT)
addHook("MobjMoveCollide", cutCollision, MT_PSYCHOCUT)

local pannel2wpn = {
	[MT_GRENADEPICKUP] = RW_GRENADE,
	[MT_BOUNCEPICKUP] = RW_BOUNCE,
	[MT_RAILPICKUP] = RW_RAIL,
	[MT_AUTOPICKUP] = RW_AUTO,
	[MT_EXPLODEPICKUP] = RW_EXPLODE,
	[MT_SCATTERPICKUP] = RW_SCATTER,
	[MT_BOUNCERING] = RW_BOUNCE,
	[MT_RAILRING] = RW_RAIL,
	[MT_BOUNCERING] = RW_BOUNCE,
	[MT_SCATTERRING] = RW_SCATTER,
	[MT_AUTOMATICRING] = RW_AUTO,
	[MT_EXPLOSIONRING] = RW_EXPLODE,
	[MT_GRENADERING] = RW_GRENADE
}
-- Pickup
local wpn2ammo = {
	[RW_AUTO] = 6,
	[RW_BOUNCE] = 6,
	[RW_SCATTER] = 2,
	[RW_GRENADE] = 2,
	[RW_EXPLODE] = 1,
	[RW_RAIL] = 3,
}
-- Panel Pickup
local wpn2maxammo = {
	[RW_AUTO] = 30,
	[RW_BOUNCE] = 12,
	[RW_SCATTER] = 10,
	[RW_GRENADE] = 8,
	[RW_EXPLODE] = 5,
	[RW_RAIL] = 9,
}

local function weaponRingsPannel(pn, mo, mo)
	if mo and mo.valid and mo.skin == "silver" and mo.player
		local wpn = pannel2wpn[pn.type]
		if not wpn return end
		mo.player.ringweapons = $|wpn
		mo.silv_ammo[wpn] = min(wpn2maxammo[wpn], $+wpn2ammo[wpn])
	end
end

local function weaponRings(pn, mo, mo)
	if mo and mo.valid and mo.skin == "silver" and mo.player
		local wpn = pannel2wpn[pn.type]
		if not wpn return end
		mo.silv_ammo[wpn] = min(wpn2maxammo[wpn], $+wpn2ammo[wpn])
	end
end

do
	for i = 0, #mobjinfo-1
		if i >= MT_BOUNCERING and i <= MT_GRENADERING
			addHook("MobjDeath", weaponRings, i)
		elseif i >= MT_BOUNCEPICKUP and i <= MT_GRENADEPICKUP
			addHook("MobjDeath", weaponRingsPannel, i)
		end
	end
end


local wep2img = {
	[RW_AUTO] = "AUTOIND",
	[RW_BOUNCE] = "BNCEIND",
	[RW_SCATTER] = "SCATIND",
	[RW_GRENADE] = "GRENIND",
	[RW_EXPLODE] = "BOMBIND",
	[RW_RAIL] = "RAILIND"
}

local wpnx = {}
for i = 0, 15
	wpnx[2^i] = 96 + (21*(i+1))
end

-- Psycho Cut HUD
rawset(_G, "psyCutHUD", function(v, p)
	if not silv_rnsHUD return end
	local mo = p.mo
	if G_RingSlingerGametype()
	and not (G_TagGametype() and not (p.pflags & PF_TAGIT))
		local transflag = ((mo.silv_telekinesis and #mo.silv_telekinesis) or mo.silv_energy < 18 or mo.silv_fire or mo.silv_firecooldown) and V_HUDTRANSHALF or 0	-- make this translucent when you can't fire
		local x, y = 96, (splitscreen) and (81+100*#p) or 162
		v.draw(x, y, v.cachePatch("ICO_CUT"), V_SNAPTOBOTTOM|transflag)
		v.draw(x-2, y-2, v.cachePatch("TL_ABL"), V_SNAPTOBOTTOM|transflag)
		if p.powers[pw_infinityring]
			v.drawString(x+8, y+8, (p.powers[pw_infinityring] >= 80 and "\x82" or "")..p.powers[pw_infinityring], V_SNAPTOBOTTOM|V_SNAPTOLEFT|transflag, "center")
		end
		-- cross combination
		for i = 0, 15 do
			local wpn = 2^i
			if mo.silv_ammo[wpn] or p.ringweapons & wpn
				local tflag = (not mo.silv_ammo[wpn] or not (p.ringweapons & wpn)) and V_60TRANS or 0
				v.draw(wpnx[wpn], y, v.cachePatch(wep2img[wpn]), V_SNAPTOBOTTOM|tflag)
				v.drawString(wpnx[wpn], y, mo.silv_ammo[wpn], V_50TRANS|V_SNAPTOBOTTOM|(mo.silv_ammo[wpn] >= wpn2maxammo[wpn] and V_YELLOWMAP or 0), string.len(tostring(mo.silv_ammo[wpn])) >= 3 and "thin" or nil)
				if p.currentweapon == i+1
					v.draw(wpnx[wpn]-2, y-2, v.cachePatch("CURWEAP"), V_SNAPTOBOTTOM)
				end
			end
		end
	end
end)